# -*- coding: utf-8 -*-
#
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import shutil
import tempfile
import unittest
import time
import json
from productmd import ComposeInfo

from mock import patch, MagicMock, mock_open, call
from kobo.conf import PyConfigParser

from odcs.server.pungi import (
    Pungi,
    PungiConfig,
    PungiSourceType,
    PungiLogs,
    RawPungiConfig,
)
from odcs.server import conf, db
from odcs.server.models import Compose
from odcs.common.types import COMPOSE_STATES, COMPOSE_RESULTS, COMPOSE_FLAGS
from odcs.server.utils import makedirs
from .utils import AnyStringWith, ModelsBaseTest

test_dir = os.path.abspath(os.path.dirname(__file__))


class TestPungiConfig(unittest.TestCase):
    def setUp(self):
        super(TestPungiConfig, self).setUp()

    def tearDown(self):
        super(TestPungiConfig, self).tearDown()

    def test_pungi_config_module(self):
        pungi_cfg = PungiConfig(
            "MBS-512",
            "1",
            PungiSourceType.MODULE,
            "testmodule:master:1:1",
            module_defaults_url="git://localhost.tld/x.git master",
        )
        cfg = pungi_cfg.get_pungi_config()
        variants = pungi_cfg.get_variants_config()
        comps = pungi_cfg.get_comps_config()

        self.assertTrue(variants.find("<module>") != -1)
        self.assertEqual(comps, "")
        self.assertEqual(
            cfg["module_defaults_dir"],
            {
                "branch": "master",
                "dir": ".",
                "repo": "git://localhost.tld/x.git",
                "scm": "git",
            },
        )

    def test_pungi_config_module_empty_source(self):
        pungi_cfg = PungiConfig(
            "MBS-512",
            "1",
            PungiSourceType.MODULE,
            "",
        )
        variants = pungi_cfg.get_variants_config()
        comps = pungi_cfg.get_comps_config()
        # No "module" in variants case the source is empty.
        self.assertTrue(variants.find("<module>") == -1)
        self.assertEqual(comps, "")

    def test_pungi_config_tag(self):
        pungi_cfg = PungiConfig(
            "MBS-512",
            "1",
            PungiSourceType.KOJI_TAG,
            "f26",
            packages=["file"],
            sigkeys="123 456",
            arches=["ppc64", "s390"],
        )
        cfg = pungi_cfg.get_pungi_config()
        variants = pungi_cfg.get_variants_config()
        comps = pungi_cfg.get_comps_config()

        self.assertTrue(variants.find("<groups>") != -1)
        self.assertTrue(variants.find("ppc64") != -1)
        self.assertTrue(variants.find("s390") != -1)
        self.assertTrue(comps.find("file</packagereq>") != -1)
        self.assertEqual(cfg["sigkeys"], ["123", "456"])

    def test_get_pungi_conf(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512", "1", PungiSourceType.MODULE, "testmodule:master:1:1"
            )
            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(cfg["release_name"], "MBS-512")
            self.assertEqual(cfg["release_short"], "MBS-512")
            self.assertEqual(cfg["release_version"], "1")
            self.assertTrue("createiso" in cfg["skip_phases"])
            self.assertTrue("buildinstall" in cfg["skip_phases"])

    @patch("odcs.server.pungi.log")
    def test_get_pungi_conf_exception(self, log):
        pungi_cfg = PungiConfig(
            "MBS-512", "1", PungiSourceType.MODULE, "testmodule:master:1:1"
        )
        _, mock_path = tempfile.mkstemp(suffix="-pungi.conf")
        with open(mock_path, "w") as f:
            # write an invalid jinja2 template file
            f.write("{{\n")
        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg.get_pungi_config()
            log.exception.assert_called_once()
        os.remove(mock_path)

    def test_get_pungi_conf_iso(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.MODULE,
                "testmodule:master:1:1",
                results=COMPOSE_RESULTS["iso"],
            )
            cfg = pungi_cfg.get_pungi_config()
            self.assertTrue("createiso" not in cfg["skip_phases"])

    def test_get_pungi_conf_boot_iso(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.MODULE,
                "testmodule:master:1:1",
                results=COMPOSE_RESULTS["boot.iso"],
            )
            cfg = pungi_cfg.get_pungi_config()
            self.assertTrue("buildinstall" not in cfg["skip_phases"])

    def test_get_pungi_conf_koji_inherit(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig("MBS-512", "1", PungiSourceType.KOJI_TAG, "f26")

            pungi_cfg.pkgset_koji_inherit = False
            cfg = pungi_cfg.get_pungi_config()
            self.assertFalse(cfg["pkgset_koji_inherit"])

            pungi_cfg.pkgset_koji_inherit = True
            cfg = pungi_cfg.get_pungi_config()
            self.assertTrue(cfg["pkgset_koji_inherit"])

    def test_get_pungi_conf_check_deps(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig("MBS-512", "1", PungiSourceType.KOJI_TAG, "f26")

            cfg = pungi_cfg.get_pungi_config()
            self.assertIs(cfg["check_deps"], False)

            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.KOJI_TAG,
                "f26",
                flags=COMPOSE_FLAGS["check_deps"],
            )
            cfg = pungi_cfg.get_pungi_config()
            self.assertIs(cfg["check_deps"], True)

    def test_get_pungi_conf_multilib(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.KOJI_TAG,
                "f26",
                multilib_arches=["x86_64", "s390x"],
                multilib_method=3,
            )

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(
                set(cfg["multilib"][0][1].keys()), set(["s390x", "x86_64"])
            )
            for variant, arch_method_dict in cfg["multilib"]:
                for method in arch_method_dict.values():
                    self.assertEqual(set(method), set(["runtime", "devel"]))

    def test_get_pungi_conf_pkgset_koji_builds(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.KOJI_TAG,
                "f26",
                builds=["foo-1-1", "bar-1-1"],
            )

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(
                set(cfg["pkgset_koji_builds"]), set(["foo-1-1", "bar-1-1"])
            )
            self.assertEqual(
                cfg["additional_packages"], [(u"^Temporary$", {u"*": [u"*"]})]
            )

    def test_get_pungi_conf_scratch_build_tasks(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.KOJI_TAG,
                "f26",
                scratch_build_tasks="123456 123457",
            )

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(
                set(cfg["pkgset_koji_scratch_tasks"]), set(["123456", "123457"])
            )

    def test_get_pungi_conf_modular_koji_tags(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.KOJI_TAG,
                "f26",
                modular_koji_tags="f26-modules",
                module_defaults_url="git://localhost.tld/x.git master",
                packages=["foo"],
            )

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(set(cfg["pkgset_koji_module_tag"]), set(["f26-modules"]))
            self.assertEqual(cfg["gather_method"], "hybrid")
            self.assertEqual(
                cfg["module_defaults_dir"],
                {
                    "branch": "master",
                    "dir": ".",
                    "repo": "git://localhost.tld/x.git",
                    "scm": "git",
                },
            )

            # The "<modules>" must appear in the variants.xml after the "<groups>".
            variants = pungi_cfg.get_variants_config()
            self.assertTrue(variants.find("<module>") != -1)
            self.assertTrue(variants.find("<groups>") != -1)
            self.assertTrue(variants.find("<module>") > variants.find("<groups>"))

    def test_get_pungi_conf_source_type_build(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.BUILD,
                "x",
                builds=["foo-1-1", "bar-1-1"],
            )

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(cfg["pkgset_koji_tag"], "")
            self.assertEqual(
                set(cfg["pkgset_koji_builds"]), set(["foo-1-1", "bar-1-1"])
            )
            self.assertEqual(
                cfg["additional_packages"], [(u"^Temporary$", {u"*": [u"*"]})]
            )

    def test_get_pungi_conf_source_type_koji_tag_all_packages(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig("MBS-512", "1", PungiSourceType.KOJI_TAG, "f26")

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(cfg["pkgset_koji_tag"], "f26")
            self.assertEqual(
                cfg["additional_packages"], [("^Temporary$", {"*": ["*"]})]
            )

    def test_get_pungi_conf_source_type_koji_tag_some_packages(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512", "1", PungiSourceType.KOJI_TAG, "f26", packages=["file"]
            )

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(cfg["pkgset_koji_tag"], "f26")
            self.assertTrue("additional_packages" not in cfg)

    def test_get_pungi_conf_lookaside_repos(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.KOJI_TAG,
                "f26",
                lookaside_repos="foo bar",
            )

            cfg = pungi_cfg.get_pungi_config()
            self.assertEqual(
                cfg["gather_lookaside_repos"], [(u"^.*$", {u"*": [u"foo", u"bar"]})]
            )

    def test_get_pungi_conf_include_devel_modules(self):
        _, mock_path = tempfile.mkstemp()
        template_path = os.path.abspath(os.path.join(test_dir, "../conf/pungi.conf"))
        shutil.copy2(template_path, mock_path)

        with patch("odcs.server.pungi.conf.pungi_conf_path", mock_path):
            pungi_cfg = PungiConfig(
                "MBS-512",
                "1",
                PungiSourceType.MODULE,
                "foo:1:1:1 foo-devel:1:1:1 bar-devel:1:1:1",
            )

            self.assertEqual(
                pungi_cfg.source, "foo:1:1:1 foo-devel:1:1:1 bar-devel:1:1:1"
            )


class FakePyConfigParser(dict):
    def load_from_file(self, *args, **kwargs):
        pass

    def load_from_string(self, *args, **kwargs):
        pass


class TestPungi(ModelsBaseTest):
    def setUp(self):
        super(TestPungi, self).setUp()

        def mocked_clone_repo(url, dest, branch="master", commit=None):
            makedirs(dest)
            makedirs(os.path.join(dest, "another"))
            with open(os.path.join(dest, "pungi.conf"), "w") as fd:
                lines = [
                    'release_name = "fake pungi conf 1"',
                    'release_short = "compose-1"',
                    'release_version = "10"',
                ]
                fd.writelines(lines)
            with open(os.path.join(dest, "another", "pungi.conf"), "w") as fd:
                lines = [
                    'release_name = "fake pungi conf 2"',
                    'release_short = "compose-2"',
                    'release_version = "10"',
                ]
                fd.writelines(lines)

        self.patch_clone_repo = patch("odcs.server.pungi.clone_repo")
        self.clone_repo = self.patch_clone_repo.start()
        self.clone_repo.side_effect = mocked_clone_repo

        self.patch_makedirs = patch("odcs.server.pungi.makedirs")
        self.makedirs = self.patch_makedirs.start()

        self.patch_ci_dump = patch("odcs.server.pungi.ComposeInfo.dump")
        self.ci_dump = self.patch_ci_dump.start()

        self.compose = MagicMock()
        self.compose.target_dir = conf.target_dir
        self.compose.toplevel_dir = os.path.join(conf.target_dir, "odcs-1")
        self.compose.compose_type = "test"
        self.compose.label = None
        self.compose.pungi_config_dump = None

        makedirs(self.compose.toplevel_dir)

    def tearDown(self):
        super(TestPungi, self).tearDown()

        self.patch_clone_repo.stop()
        self.patch_makedirs.stop()
        self.patch_ci_dump.stop()

        shutil.rmtree(self.compose.toplevel_dir)

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run(self, execute_cmd):
        pungi_cfg = PungiConfig(
            "MBS-512", "1", PungiSourceType.MODULE, "testmodule:master:1:1"
        )
        pungi = Pungi(1, pungi_cfg)
        pungi.run(self.compose)

        self.makedirs.assert_called_with(AnyStringWith("test_composes/odcs-1/"))
        self.makedirs.assert_called_with(AnyStringWith("work/global"))
        self.ci_dump.assert_called_once_with(
            AnyStringWith("work/global/composeinfo-base.json")
        )

        execute_cmd.assert_called_once_with(
            [
                "pungi-koji",
                AnyStringWith("pungi.json"),
                "--no-latest-link",
                AnyStringWith("--compose-dir="),
                "--test",
            ],
            cwd=AnyStringWith("/tmp/"),
            timeout=3600,
            stderr=AnyStringWith("pungi-stderr.log"),
            stdout=AnyStringWith("pungi-stdout.log"),
        )

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run_parent_pungi_compose_ids(self, execute_cmd):
        self.compose.parent_pungi_compose_ids = "Fedora-1-1 Fedora-2-2"
        pungi_cfg = PungiConfig(
            "MBS-512", "1", PungiSourceType.MODULE, "testmodule:master:1:1"
        )
        pungi = Pungi(1, pungi_cfg)
        pungi.run(self.compose)

        self.makedirs.assert_called_with(AnyStringWith("test_composes/odcs-1/"))
        self.makedirs.assert_called_with(AnyStringWith("work/global"))
        self.ci_dump.assert_called_once_with(
            AnyStringWith("work/global/composeinfo-base.json")
        )

        execute_cmd.assert_called_once_with(
            [
                "pungi-koji",
                AnyStringWith("pungi.json"),
                "--no-latest-link",
                AnyStringWith("--compose-dir="),
                "--test",
                "--parent-compose-id=Fedora-1-1",
                "--parent-compose-id=Fedora-2-2",
            ],
            cwd=AnyStringWith("/tmp/"),
            timeout=3600,
            stderr=AnyStringWith("pungi-stderr.log"),
            stdout=AnyStringWith("pungi-stdout.log"),
        )

    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.pungi.PyConfigParser")
    def test_pungi_run_cts(self, py_config_parser, execute_cmd):
        self.patch_ci_dump.stop()
        py_config_parser.return_value = FakePyConfigParser(
            {"cts_url": "https://cts.localhost.tld/", "cts_keytab": "/tmp/some.keytab"}
        )

        def fake_execute_cmd(*args, **kwargs):
            # Fake `execute_cmd` method which creates composeinfo-base.json file
            # and waits for three seconds to test that ODCS picks up the compose
            # ID from this file.
            p = os.path.join(
                self.compose.toplevel_dir, "work", "global", "composeinfo-base.json"
            )
            makedirs(os.path.dirname(p))
            ci = ComposeInfo()
            ci.compose.id = "Fedora-Rawhide-20200517.n.1"
            ci.compose.type = "nightly"
            ci.compose.date = "20200517"
            ci.compose.respin = 1
            ci.release.name = "Fedora"
            ci.release.short = "Fedora"
            ci.release.version = "Rawhide"
            ci.release.is_layered = False
            ci.release.type = "ga"
            ci.release.internal = False
            ci.dump(p)
            time.sleep(3)

        execute_cmd.side_effect = fake_execute_cmd

        pungi_cfg = PungiConfig(
            "MBS-512", "1", PungiSourceType.MODULE, "testmodule:master:1:1"
        )
        pungi = Pungi(1, pungi_cfg)
        pungi.run(self.compose)

        self.makedirs.assert_called_with(AnyStringWith("test_composes/odcs-1"))

        execute_cmd.assert_called_once_with(
            [
                "pungi-koji",
                AnyStringWith("pungi.json"),
                "--no-latest-link",
                AnyStringWith("--compose-dir="),
                "--test",
            ],
            cwd=AnyStringWith("/tmp/"),
            timeout=3600,
            stderr=AnyStringWith("pungi-stderr.log"),
            stdout=AnyStringWith("pungi-stdout.log"),
        )

        self.assertEqual(self.compose.pungi_compose_id, "Fedora-Rawhide-20200517.n.1")

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run_compose_type(self, execute_cmd):
        for compose_type in [None, "test", "ci", "nightly", "production"]:
            self.makedirs.reset_mock()
            self.ci_dump.reset_mock()
            execute_cmd.reset_mock()

            self.compose.compose_type = compose_type
            pungi_cfg = PungiConfig(
                "MBS-512", "1", PungiSourceType.MODULE, "testmodule:master:1:1"
            )
            pungi = Pungi(1, pungi_cfg)
            pungi.run(self.compose)

            self.makedirs.assert_called_with(AnyStringWith("test_composes/odcs-1/"))
            self.makedirs.assert_called_with(AnyStringWith("work/global"))
            self.ci_dump.assert_called_once_with(
                AnyStringWith("work/global/composeinfo-base.json")
            )

            execute_cmd.assert_called_once_with(
                [
                    "pungi-koji",
                    AnyStringWith("pungi.json"),
                    "--no-latest-link",
                    AnyStringWith("--compose-dir="),
                    "--%s" % (compose_type or "test"),
                ],
                cwd=AnyStringWith("/tmp/"),
                timeout=3600,
                stderr=AnyStringWith("pungi-stderr.log"),
                stdout=AnyStringWith("pungi-stdout.log"),
            )

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run_raw_config(self, execute_cmd):
        def mocked_execute_cmd(*args, **kwargs):
            topdir = kwargs["cwd"]
            with open(os.path.join(topdir, "pungi.conf"), "r") as f:
                data = f.read()
                self.assertTrue("fake pungi conf 1" in data)

        execute_cmd.side_effect = mocked_execute_cmd

        fake_raw_config_urls = {
            "pungi.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "pungi.conf",
            }
        }
        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            self.compose.source = "pungi.conf#hash"
            pungi = Pungi(1, RawPungiConfig(self.compose))
            pungi.run(self.compose)

        self.makedirs.assert_called_with(AnyStringWith("test_composes/odcs-1/"))
        self.makedirs.assert_called_with(AnyStringWith("work/global"))
        self.ci_dump.assert_called_once_with(
            AnyStringWith("work/global/composeinfo-base.json")
        )

        execute_cmd.assert_called_once()
        self.clone_repo.assert_called_once_with(
            "http://localhost/test.git",
            AnyStringWith("/raw_config_repo"),
            commit="hash",
        )
        compose_date = time.strftime("%Y%m%d", time.localtime())
        self.assertEqual(
            self.compose.pungi_compose_id, "compose-1-10-%s.t.0" % compose_date
        )

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run_raw_config_respin(self, execute_cmd):
        compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.RAW_CONFIG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(compose)
        db.session.commit()

        def mocked_execute_cmd(*args, **kwargs):
            topdir = kwargs["cwd"]
            with open(os.path.join(topdir, "pungi.conf"), "r") as f:
                data = f.read()
                self.assertTrue("fake pungi conf 1" in data)

        execute_cmd.side_effect = mocked_execute_cmd

        fake_raw_config_urls = {
            "pungi.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "pungi.conf",
            }
        }
        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            self.compose.source = "pungi.conf#hash"
            pungi = Pungi(1, RawPungiConfig(self.compose))
            pungi.run(compose)
            pungi.run(compose)

        compose_date = time.strftime("%Y%m%d", time.localtime())
        self.assertEqual(compose.pungi_compose_id, "compose-1-10-%s.t.1" % compose_date)

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run_raw_config_subpath(self, execute_cmd):
        def mocked_execute_cmd(*args, **kwargs):
            topdir = kwargs["cwd"]
            with open(os.path.join(topdir, "pungi.conf"), "r") as f:
                data = f.read()
                self.assertTrue("fake pungi conf 2" in data)

        execute_cmd.side_effect = mocked_execute_cmd

        fake_raw_config_urls = {
            "pungi.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "pungi.conf",
                "path": "another",
            }
        }
        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            self.compose.source = "pungi.conf#hash"
            pungi = Pungi(1, RawPungiConfig(self.compose))
            pungi.run(self.compose)

        execute_cmd.assert_called_once()
        self.clone_repo.assert_called_once_with(
            "http://localhost/test.git",
            AnyStringWith("/raw_config_repo"),
            commit="hash",
        )

    @patch("odcs.server.utils.execute_cmd")
    def test_raw_config_validate(self, execute_cmd):
        fake_raw_config_urls = {
            "pungi.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "pungi.conf",
                "schema_override": "/etc/odcs/extra_override.json",
            }
        }
        with patch.object(
            conf, "raw_config_schema_override", new="/etc/odcs/default_override.json"
        ):
            with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
                with patch.object(
                    conf, "pungi_config_validate", new="pungi-config-validate"
                ):
                    self.compose.source = "pungi.conf#hash"
                    pungi = Pungi(1, RawPungiConfig(self.compose))
                    pungi.run(self.compose)

        self.assertEqual(
            execute_cmd.mock_calls[0],
            call(
                [
                    "pungi-config-validate",
                    "--old-composes",
                    "--schema-override",
                    "/etc/odcs/default_override.json",
                    "--schema-override",
                    "/etc/odcs/extra_override.json",
                    AnyStringWith("pungi.json"),
                ],
                stderr=AnyStringWith("pungi-config-validate-stderr.log"),
                stdout=AnyStringWith("pungi-config-validate-stdout.log"),
            ),
        )

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run_raw_config_custom_timeout(self, execute_cmd):
        fake_raw_config_urls = {
            "pungi.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "pungi.conf",
                "pungi_timeout": 7200,
            }
        }
        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            self.compose.source = "pungi.conf#hash"
            pungi = Pungi(1, RawPungiConfig(self.compose))
            pungi.run(self.compose)

        execute_cmd.assert_called_once_with(
            [
                "pungi-koji",
                AnyStringWith("pungi.json"),
                "--no-latest-link",
                AnyStringWith("--compose-dir="),
                "--test",
            ],
            cwd=AnyStringWith("/tmp/"),
            timeout=7200,
            stderr=AnyStringWith("pungi-stderr.log"),
            stdout=AnyStringWith("pungi-stdout.log"),
        )

    @patch("odcs.server.utils.execute_cmd")
    def test_pungi_run_raw_config_label(self, execute_cmd):
        self.compose.label = "Alpha-0.1"

        fake_raw_config_urls = {
            "pungi.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "pungi.conf",
                "pungi_timeout": 7200,
            }
        }
        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            self.compose.source = "pungi.conf#hash"
            pungi = Pungi(1, RawPungiConfig(self.compose))
            pungi.run(self.compose)

        execute_cmd.assert_called_once_with(
            [
                "pungi-koji",
                AnyStringWith("pungi.json"),
                "--no-latest-link",
                AnyStringWith("--compose-dir="),
                "--test",
                "--label=Alpha-0.1",
            ],
            cwd=AnyStringWith("/tmp/"),
            timeout=7200,
            stderr=AnyStringWith("pungi-stderr.log"),
            stdout=AnyStringWith("pungi-stdout.log"),
        )


class TestRawPungiConfig(ModelsBaseTest):
    def setUp(self):
        super(TestRawPungiConfig, self).setUp()
        self.compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.RAW_CONFIG,
            "test.conf#hash",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        def mocked_clone_repo(url, dest, branch="master", commit=None):
            makedirs(dest)
            with open(os.path.join(dest, "test.conf"), "w") as fd:
                lines = [
                    'release_name = "fake pungi conf 1"',
                    'release_short = "compose-1"',
                    'release_version = "10"',
                ]
                fd.writelines(lines)

        self.patch_clone_repo = patch("odcs.server.pungi.clone_repo")
        self.clone_repo = self.patch_clone_repo.start()
        self.clone_repo.side_effect = mocked_clone_repo

    def tearDown(self):
        super(TestRawPungiConfig, self).tearDown()
        self.patch_clone_repo.stop()

    def test_raw_config_custom_wrapper(self):
        custom_raw_config_wrapper = os.path.join(
            test_dir, "data", "custom_raw_config_wrapper.conf"
        )
        fake_raw_config_urls = {
            "test.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "test.conf",
                "raw_config_wrapper": custom_raw_config_wrapper,
            }
        }
        self.compose.builds = "foo-1-1 bar-1-1"

        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            td = tempfile.mkdtemp()
            try:
                cfg = RawPungiConfig(self.compose)
                cfg.write_config_files(td)
                pungi_conf = PyConfigParser()
                pungi_conf.load_from_file(os.path.join(td, "pungi.conf"))
                self.assertEqual(pungi_conf["release_short"], "compose-1")
                self.assertEqual(
                    pungi_conf["pkgset_koji_builds"], ["foo-1-1", "bar-1-1"]
                )
            finally:
                shutil.rmtree(td)

    def test_raw_config_pungi_config_dump(self):
        custom_raw_config_wrapper = os.path.join(
            test_dir, "data", "custom_raw_config_wrapper.conf"
        )
        fake_raw_config_urls = {
            "test.conf": {
                "url": "http://localhost/test.git",
                "config_filename": "test.conf",
                "raw_config_wrapper": custom_raw_config_wrapper,
            }
        }
        config = {
            "release_name": "foo",
            "release_short": "compose-2",
            "release_version": 50,
            "gather_lookaside_repos": [["^Buildroot$", {"aarch64": ["foo", "bar"]}]],
            "sigkeys": ["foo", None],
        }
        self.compose.pungi_config_dump = json.dumps(config)
        self.compose.builds = "foo-1-1 bar-1-1"

        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            td = tempfile.mkdtemp()
            try:
                cfg = RawPungiConfig(self.compose)
                cfg.write_config_files(td)
                with open(os.path.join(td, "pungi.json")) as fd:
                    pungi_conf = json.load(fd)
                for key in config.keys():
                    self.assertEqual(pungi_conf[key], config[key])
            finally:
                shutil.rmtree(td)


class TestPungiLogs(ModelsBaseTest):
    def setUp(self):
        super(TestPungiLogs, self).setUp()
        self.compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "tag",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        self.compose.state = COMPOSE_STATES["failed"]
        db.session.add(self.compose)
        db.session.commit()

    def tearDown(self):
        super(TestPungiLogs, self).tearDown()

    @patch("odcs.server.pungi.open", create=True)
    def test_error_string(self, patched_open):
        pungi_log = """
2018-03-23 03:38:42 [INFO    ] Writing pungi config
2018-03-23 03:38:42 [INFO    ] [BEGIN] Running pungi
2018-03-22 17:10:49 [ERROR   ] Compose run failed: No such entry in table tag: tag
2018-03-23 03:38:42 [ERROR   ] Compose run failed: ERROR running command: pungi -G
For more details see {0}/odcs-717-1-20180323.n.0/work/x86_64/pungi/Temporary.x86_64.log
2018-03-23 03:38:42 [ERROR   ] Extended traceback in: {0}/odcs-717-1-20180323.n.0/logs/global/traceback.global.log
2018-03-23 03:38:42 [CRITICAL] Compose failed: {0}/odcs-717-1-20180323.n.0
        """.format(
            conf.target_dir
        )
        patched_open.return_value = mock_open(read_data=pungi_log).return_value

        pungi_logs = PungiLogs(self.compose)
        errors = pungi_logs.get_error_string()
        self.assertEqual(
            errors,
            "Compose run failed: No such entry in table tag: tag\n"
            "Compose run failed: ERROR running command: pungi -G\n"
            "For more details see http://localhost/odcs/odcs-717-1-20180323.n.0/work/x86_64/pungi/Temporary.x86_64.log\n",
        )

    @patch("odcs.server.pungi.open", create=True)
    def test_error_string_too_many_errors(self, patched_open):
        pungi_log = (
            """
2018-03-23 03:38:42 [INFO    ] Writing pungi config
2018-03-22 17:10:49 [ERROR   ] Compose run failed: No such entry in table tag: tag
        """
            * 100
        )
        patched_open.return_value = mock_open(read_data=pungi_log).return_value

        pungi_logs = PungiLogs(self.compose)
        errors = pungi_logs.get_error_string()
        self.assertTrue("Too many errors" in errors)
        self.assertEqual(len(errors), 2058)

    @patch("odcs.server.pungi.open", create=True)
    def test_error_string_no_error(self, patched_open):
        pungi_log = """
2018-03-23 03:38:42 [INFO    ] Writing pungi config
2018-03-23 03:38:42 [INFO    ] [BEGIN] Running pungi
        """
        patched_open.return_value = mock_open(read_data=pungi_log).return_value

        pungi_logs = PungiLogs(self.compose)
        errors = pungi_logs.get_error_string()
        self.assertEqual(errors, "")

    def test_error_string_no_log(self):
        pungi_logs = PungiLogs(self.compose)
        errors = pungi_logs.get_error_string()
        self.assertEqual(errors, "")

    def test_toplevel_work_dir(self):
        # The self.glob is inherited from ModelsBaseTest.
        pungi_logs = PungiLogs(self.compose)
        errors = pungi_logs.get_error_string()
        self.assertEqual(errors, "")

    @patch("odcs.server.pungi.open", create=True)
    def test_config_dump(self, patched_open):
        patched_open.return_value = mock_open(
            read_data="fake\npungi\nconf\n"
        ).return_value

        pungi_logs = PungiLogs(self.compose)
        ret = pungi_logs.get_config_dump()
        self.assertEqual(ret, "fake\npungi\nconf\n")

        patched_open.assert_called_once_with(
            AnyStringWith("logs/global/config-dump.global.log"), "r"
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Jan Kaluza <jkaluza@redhat.com>

import copy
import json
import re
import requests

from odcs.server import conf, log
from odcs.server.mergerepo import MergeRepo
from odcs.server.utils import retry


class Pulp(object):
    """Interface to Pulp"""

    def __init__(self, server_url, username, password, compose):
        self.username = username
        self.password = password
        self.server_url = server_url
        self.compose = compose
        self.rest_api_root = "{0}/pulp/api/v2/".format(self.server_url.rstrip("/"))

    @retry(wait_on=requests.exceptions.RequestException)
    def _rest_post(self, endpoint, post_data):
        query_data = json.dumps(post_data)
        try:
            r = requests.post(
                "{0}{1}".format(self.rest_api_root, endpoint.lstrip("/")),
                query_data,
                auth=(self.username, self.password),
                timeout=conf.net_timeout,
            )
        except requests.exceptions.RequestException as e:
            # also catches ConnectTimeout, ConnectionError
            # change message of the catched exception and re-raise
            msg = "Pulp connection has failed: {}".format(e.args)
            raise requests.exceptions.RequestException(msg)

        r.raise_for_status()
        return r.json()

    def _try_arch_merge(self, content_set_repos):
        """
        Tries replacing arch string (e.g. "x86_64" or "ppc64le") in each "url"
        in content_set_repos with "$basearch" and if this results in the same
        repository URL for each repo in the content_set_repos and also
        the sigkeys are the same, returns the single repo with $basearch.
        If not, returns an empty dict.

        The "arches" value of returned repo is set to union of merged "arches".

        For example, for following input:
            [{"url": "http://localhost/x86_64/os", "arches": ["x86_64"]},
             {"url": "http://localhost/ppc64le/os", "arches": ["ppc64le"]}]
        This method returns:
            {"url": "http://localhost/$basearch/os",
             "arches": ["x86_64", "ppc64le"]}
        """
        # For no or exactly one repo, there is nothing to merge.
        if len(content_set_repos) < 2:
            return {}

        first_repo = None
        for repo in content_set_repos:
            if len(repo["arches"]) != 1:
                # This should not happen normally, because each repo has just
                # single arch in Pulp, but be defensive.
                raise ValueError(
                    "Content set repository %s does not have exactly 1 arch: "
                    "%r." % (repo["url"], repo["arches"])
                )
            url = repo["url"].replace(list(repo["arches"])[0], "$basearch")
            if first_repo is None:
                first_repo = copy.deepcopy(repo)
                first_repo["url"] = url
                continue
            if first_repo["url"] != url or first_repo["sigkeys"] != repo["sigkeys"]:
                return {}
            first_repo["arches"] = first_repo["arches"].union(repo["arches"])
        return first_repo

    def _merge_repos(self, content_set, content_set_repos):
        """
        Merges the repositories of the same arch from `content_set_repos`
        and returns the new repository dict pointing to the newly created
        merged repository.

        In case there is just single (or none) repository in
        `content_set_repos`, returns empty dict.
        """
        # For no or exactly one repo, there is nothing to merge.
        if len(content_set_repos) < 2:
            return {}

        # We must merge repos of the same arch only, so group them by arch
        # at first.
        per_arch_repos = {}
        for repo in content_set_repos:
            if len(repo["arches"]) != 1:
                # This should not happen normally, because each repo has just
                # single arch in Pulp, but be defensive.
                raise ValueError(
                    "Content set repository %s does not have exactly 1 arch: "
                    "%r." % (repo["url"], repo["arches"])
                )
            arch = list(repo["arches"])[0]
            if arch not in per_arch_repos:
                per_arch_repos[arch] = []
            per_arch_repos[arch].append(repo)

        merge_repo = MergeRepo(self.compose)

        for arch, repos in per_arch_repos.items():
            urls = [repo["url"] for repo in repos]
            merge_repo.run(arch, urls, content_set)

        return {
            "url": "%s/%s/$basearch" % (self.compose.result_repo_url, content_set),
            "arches": set(per_arch_repos.keys()),
            "sigkeys": content_set_repos[0]["sigkeys"],
        }

    def get_repos_from_content_sets(
        self, content_sets, include_unpublished_repos=False
    ):
        """
        Returns dictionary with URLs of all shipped repositories defined by
        the content_sets.
        The key in the returned dict is the content_set name and the value
        is the URL to repository with RPMs.

        :param list content_sets: Content sets to look for.
        :rtype: dict
        :return: Dictionary in following format:
            {
                content_set_1: {
                    "url": repo_url,
                    "arches": set([repo_arch1, repo_arch2]),
                    'sigkeys': ['sigkey1', 'sigkey2', ...]
                },
                ...
            }
        """
        query_data = {
            "criteria": {
                "filters": {"notes.content_set": {"$in": content_sets}},
                "fields": ["notes"],
            }
        }

        if not include_unpublished_repos:
            query_data["criteria"]["filters"][
                "notes.include_in_download_service"
            ] = "True"
        repos = self._rest_post("repositories/search/", query_data)

        per_content_set_repos = {}
        for repo in repos:
            notes = repo["notes"]
            url = "%s/%s" % (self.server_url.rstrip("/"), notes["relative_url"])
            arch = notes["arch"]
            sigkeys = sorted(notes["signatures"].split(","))
            # OSBS cannot verify https during the container image build, so
            # fallback to http for now.
            if url.startswith("https://"):
                url = "http://" + url[len("https://") :]
            if notes["content_set"] not in per_content_set_repos:
                per_content_set_repos[notes["content_set"]] = []
            per_content_set_repos[notes["content_set"]].append(
                {
                    "url": url,
                    "arches": set([arch]),
                    "sigkeys": sigkeys,
                    "product_versions": notes["product_versions"],
                }
            )

        ret = {}
        for cs, repos in per_content_set_repos.items():
            can_sort = True
            version_len = None
            arches = None
            for repo in repos:
                if arches and arches != repo["arches"]:
                    log.debug("Skipping sorting repos: arch mismatch.")
                    can_sort = False
                    break
                arches = repo["arches"]
                try:
                    product_versions = json.loads(repo["product_versions"])
                except ValueError:
                    log.debug(
                        "Skipping sorting repos: invalid JSON in product_versions."
                    )
                    can_sort = False
                    break
                if len(product_versions) != 1:
                    log.debug(
                        "Skipping sorting repos: more than one version for a repo."
                    )
                    can_sort = False
                    break
                if not re.match(r"\d+(\.\d+)*$", product_versions[0]):
                    log.debug("Skipping sorting repos: unparsable version string.")
                    can_sort = False
                    break
                # Parse the version into a tuple.
                ver_len = product_versions[0].count(".")
                if version_len and version_len != ver_len:
                    log.debug("Skipping sorting repos: incompatible versions.")
                    can_sort = False
                    break
                version_len = ver_len
            if can_sort:
                repos = [
                    sorted(
                        repos,
                        key=lambda k: tuple(
                            json.loads(k["product_versions"])[0].split(".")
                        ),
                        reverse=True,
                    )[0]
                ]
            # In case there are multiple repos, at first try merging them
            # by replacing arch in repository baseurl with $basearch.
            merged_repos = self._try_arch_merge(repos)
            if not merged_repos:
                # In case we cannot merge repositories by replacing the arch
                # with $basearch, call mergerepo_c.
                merged_repos = self._merge_repos(cs, repos)
            if merged_repos:
                ret[cs] = merged_repos
            else:
                ret[cs] = repos[-1]

        return ret



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

from mock import patch

from odcs.server.pulp import Pulp
from odcs.server.pungi import PungiSourceType
from odcs.server import db, conf
from odcs.server.models import Compose

from .utils import ModelsBaseTest


@patch("odcs.server.pulp.Pulp._rest_post")
class TestPulp(ModelsBaseTest):
    def test_pulp_request(self, pulp_rest_post):
        c = Compose.create(db.session, "me", PungiSourceType.PULP, "foo-1", 0, 3600)
        db.session.commit()

        pulp_rest_post.return_value = []

        pulp = Pulp("http://localhost/", "user", "pass", c)
        pulp.get_repos_from_content_sets(["foo-1", "foo-2"])
        pulp_rest_post.assert_called_once_with(
            "repositories/search/",
            {
                "criteria": {
                    "fields": ["notes"],
                    "filters": {
                        "notes.include_in_download_service": "True",
                        "notes.content_set": {"$in": ["foo-1", "foo-2"]},
                    },
                }
            },
        )

    def test_pulp_request_include_inpublished(self, pulp_rest_post):
        c = Compose.create(db.session, "me", PungiSourceType.PULP, "foo-1", 0, 3600)
        db.session.commit()

        pulp_rest_post.return_value = []

        pulp = Pulp("http://localhost/", "user", "pass", c)
        pulp.get_repos_from_content_sets(["foo-1", "foo-2"], True)
        pulp_rest_post.assert_called_once_with(
            "repositories/search/",
            {
                "criteria": {
                    "fields": ["notes"],
                    "filters": {"notes.content_set": {"$in": ["foo-1", "foo-2"]}},
                }
            },
        )

    def test_generate_pulp_compose_arch_merge(self, pulp_rest_post):
        """
        Tests that multiple repos in single content_set are merged into
        single one by replacing arch with $basearch variable if possible.
        """
        c = Compose.create(db.session, "me", PungiSourceType.PULP, "foo-1", 0, 3600)
        db.session.commit()

        pulp_rest_post.return_value = [
            {
                "notes": {
                    "relative_url": "content/1/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                },
            },
            {
                "notes": {
                    "relative_url": "content/1/ppc64le/os",
                    "content_set": "foo-1",
                    "arch": "ppc64le",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                }
            },
            {
                "notes": {
                    "relative_url": "content/3/ppc64/os",
                    "content_set": "foo-2",
                    "arch": "ppc64",
                    "signatures": "SIG1,SIG3",
                    "product_versions": "",
                }
            },
        ]

        pulp = Pulp("http://localhost/", "user", "pass", c)
        ret = pulp.get_repos_from_content_sets(["foo-1", "foo-2"])
        self.assertEqual(
            ret,
            {
                "foo-1": {
                    "url": "http://localhost/content/1/$basearch/os",
                    "arches": set(["x86_64", "ppc64le"]),
                    "sigkeys": ["SIG1", "SIG2"],
                    "product_versions": "",
                },
                "foo-2": {
                    "url": "http://localhost/content/3/ppc64/os",
                    "arches": set(["ppc64"]),
                    "sigkeys": ["SIG1", "SIG3"],
                    "product_versions": "",
                },
            },
        )

    @patch("odcs.server.mergerepo.execute_cmd")
    @patch("odcs.server.mergerepo.makedirs")
    @patch("odcs.server.mergerepo.Lock")
    @patch("odcs.server.mergerepo.MergeRepo._download_repodata")
    def test_pulp_compose_merge_repos(
        self, download_repodata, lock, makedirs, execute_cmd, pulp_rest_post
    ):
        c = Compose.create(db.session, "me", PungiSourceType.PULP, "foo-1", 0, 3600)
        db.session.commit()

        pulp_rest_post.return_value = [
            {
                "notes": {
                    "relative_url": "content/1.0/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                },
            },
            # Test two same relative_urls here.
            {
                "notes": {
                    "relative_url": "content/1.0/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                },
            },
            {
                "notes": {
                    "relative_url": "content/1.1/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                }
            },
            {
                "notes": {
                    "relative_url": "content/1.0/ppc64le/os",
                    "content_set": "foo-1",
                    "arch": "ppc64le",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                },
            },
        ]

        pulp = Pulp("http://localhost/", "user", "pass", c)
        ret = pulp.get_repos_from_content_sets(["foo-1", "foo-2"])

        self.assertEqual(
            ret,
            {
                "foo-1": {
                    "url": "http://localhost/odcs/odcs-1/compose/Temporary/foo-1/$basearch",
                    "arches": set(["x86_64", "ppc64le"]),
                    "sigkeys": ["SIG1", "SIG2"],
                }
            },
        )

        makedirs.assert_any_call(c.result_repo_dir + "/foo-1/x86_64")
        makedirs.assert_any_call(c.result_repo_dir + "/foo-1/ppc64le")

        repo_prefix = "%s/pulp_repo_cache/content/" % conf.target_dir
        execute_cmd.assert_any_call(
            [
                "/usr/bin/mergerepo_c",
                "--method",
                "nvr",
                "-o",
                c.result_repo_dir + "/foo-1/x86_64",
                "--repo-prefix-search",
                "%s/pulp_repo_cache" % conf.target_dir,
                "--repo-prefix-replace",
                "http://localhost/",
                "-r",
                repo_prefix + "1.0/x86_64/os",
                "-r",
                repo_prefix + "1.1/x86_64/os",
            ],
            timeout=1800,
        )
        execute_cmd.assert_any_call(
            [
                "/usr/bin/mergerepo_c",
                "--method",
                "nvr",
                "-o",
                c.result_repo_dir + "/foo-1/ppc64le",
                "--repo-prefix-search",
                "%s/pulp_repo_cache" % conf.target_dir,
                "--repo-prefix-replace",
                "http://localhost/",
                "-r",
                repo_prefix + "1.0/ppc64le/os",
            ],
            timeout=1800,
        )

        download_repodata.assert_any_call(
            repo_prefix + "1.0/x86_64/os", "http://localhost/content/1.0/x86_64/os"
        )
        download_repodata.assert_any_call(
            repo_prefix + "1.1/x86_64/os", "http://localhost/content/1.1/x86_64/os"
        )
        download_repodata.assert_any_call(
            repo_prefix + "1.0/ppc64le/os", "http://localhost/content/1.0/ppc64le/os"
        )

    @patch("odcs.server.mergerepo.execute_cmd")
    @patch("odcs.server.mergerepo.makedirs")
    @patch("odcs.server.mergerepo.Lock")
    @patch("odcs.server.mergerepo.MergeRepo._download_repodata")
    def test_pulp_compose_find_latest_version(
        self, download_repodata, lock, makedirs, execute_cmd, pulp_rest_post
    ):
        c = Compose.create(db.session, "me", PungiSourceType.PULP, "foo-1", 0, 3600)
        db.session.commit()

        pulp_rest_post.return_value = [
            {
                "notes": {
                    "relative_url": "content/1.0/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": '["1.0"]',
                },
            },
            {
                "notes": {
                    "relative_url": "content/1.1/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": '["1.1"]',
                }
            },
        ]

        pulp = Pulp("http://localhost/", "user", "pass", c)
        ret = pulp.get_repos_from_content_sets(["foo-1"])

        self.assertEqual(
            ret,
            {
                "foo-1": {
                    "url": "http://localhost/content/1.1/x86_64/os",
                    "arches": set(["x86_64"]),
                    "sigkeys": ["SIG1", "SIG2"],
                    "product_versions": '["1.1"]',
                }
            },
        )

        makedirs.assert_not_called()

        execute_cmd.assert_not_called()

        download_repodata.assert_not_called()



"""Add Compose.scratch_modules

Revision ID: 512890e6864d
Revises: 812f2745248f
Create Date: 2020-07-31 22:40:13.138130

"""

# revision identifiers, used by Alembic.
revision = "512890e6864d"
down_revision = "812f2745248f"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("composes", sa.Column("scratch_modules", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("composes", "scratch_modules")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import requests
from collections import defaultdict

from odcs.server.utils import retry, to_text_type
from odcs.server import log

import gi

gi.require_version("Modulemd", "2.0")
from gi.repository import Modulemd  # noqa: E402


class ModuleLookupError(Exception):
    pass


class MBS(object):
    def __init__(self, config):
        self.mbs_url = config.mbs_url.rstrip("/")

    @retry(wait_on=(requests.ConnectionError,), logger=log)
    def get_modules(self, **params):
        url = self.mbs_url + "/1/module-builds/"
        r = requests.get(url, params=params)
        r.raise_for_status()
        return r.json()

    def get_latest_modules(self, nsvc, include_done=False):
        """
        Query MBS and return the latest version of the module specified by nsvc.

        :param nsvc: N:S:V[:C] of a module to include in a compose.
        :param include_done: When True, also module builds in the "done" state
            are included in a result.
        :raises ModuleLookupError: if the module couldn't be found
        :return: the latest version of the module.
        """
        if not nsvc:
            # Return an empty list of no NSVC is defined.
            return []

        params = {
            "nsvc": nsvc,
            "state": [3, 5] if include_done else [5],  # 5 is "ready", 3 is "done".
            "verbose": True,  # Needed to get modulemd in response.
            "order_desc_by": "version",
        }
        modules = self.get_modules(**params)

        # True if the module is "-devel" module.
        devel_module = False
        if not modules["meta"]["total"]:
            # In case this is "-devel" module, it won't be included in MBS,
            # but it exists as CG build in Koji. It therefore can be used
            # to generate the compose, but to actually find the MBS build,
            # we need to remove the "-devel" suffix from the NSVC.
            n = nsvc.split(":")[0]
            if n.endswith("-devel"):
                params["nsvc"] = n[: -len("-devel")] + params["nsvc"][len(n) :]
                modules = self.get_modules(**params)
                devel_module = True

        if not modules["meta"]["total"]:
            state_msg = "ready or done" if include_done else "ready"
            raise ModuleLookupError(
                "Failed to find module %s in %s state in the MBS." % (nsvc, state_msg)
            )

        ret = []
        # In case the nsvc is just "name:stream", there might be multiple
        # versions of a module in MBS response. The modules in response are
        # sorted DESC by version, so the latest module is always the first
        # one. So simply get the first module and add to `ret` all the next
        # modules in a response list which have the same version - this
        # basically adds all the contexts of the module with latest version
        # to `ret`.
        for module in modules["items"]:
            if ret and ret[0]["version"] != module["version"]:
                break
            if devel_module:
                # Add -devel to module metadata in case we are composing devel
                # module.
                module["name"] += "-devel"
                # Devel module always depend on the non-devel version
                mmd = Modulemd.ModuleStream.read_string(
                    module["modulemd"],
                    strict=True,
                    module_name=None,
                    module_stream=None,
                )
                mmd = mmd.upgrade(2)
                for dep in mmd.get_dependencies():
                    dep.add_runtime_stream(mmd.get_module_name(), mmd.get_stream_name())
                mod_index = Modulemd.ModuleIndex.new()
                mod_index.add_module_stream(mmd)
                module["modulemd"] = to_text_type(mod_index.dump_to_string())
            ret.append(module)
        return ret

    def _add_new_dependencies(self, module_map, modules):
        """
        Helper for ``validate_module_list()`` - scans ``modules`` and adds any missing
        requirements to ``module_map``.

        :param module_map: dict mapping module name:stream to module.
        :param modules: the list of modules to scan for dependencies.
        :return: a list of any modules that were added to ``module_map``.
        """

        new_modules = []
        for module in modules:
            mmd = Modulemd.ModuleStream.read_string(
                module["modulemd"], strict=True, module_name=None, module_stream=None
            )
            mmd = mmd.upgrade(2)

            # Check runtime dependency (name:stream) of a module and if this
            # dependency is already in module_map/new_modules, do nothing.
            # But otherwise get the latest module in this name:stream from MBS
            # and add it to new_modules/module_map.
            for deps in mmd.get_dependencies():
                for name in deps.get_runtime_modules():
                    for stream in deps.get_runtime_streams(name):
                        key = "%s:%s" % (name, stream)
                        if key not in module_map:
                            new_module = self.get_latest_modules(key)
                            new_modules += new_module
                            module_map[key] = [new_modules]

        return new_modules

    def validate_module_list(self, modules, expand=True):
        """
        Given a list of modules as returned by `get_modules()`, checks that
        there are no conflicting duplicates, removes any exact duplicates,
        and if ``expand`` is set, recursively adds in required modules until
        all dependencies are specified.

        :param modules: a list of modules as returned by ``get_modules()`` or
                ``get_latest_module()``
        :param expand: if required modules should be included in the returned
                list.
        :return: the list of modules with deduplication and expansion.
        :raises ModuleLookupError: if a required module couldn't be found, or a
                conflict occurred when resolving dependencies.
        """

        # List of modules we are going to return.
        new_modules = []
        # Temporary dict with "name:stream" as key and list of module dicts
        # as value.
        module_map = defaultdict(list)

        for module in modules:
            key = "%s:%s" % (module["name"], module["stream"])

            # In case this is the first module with this name:stream,
            # just add it to new_modules.
            old_modules = module_map[key]
            if not old_modules:
                module_map[key].append(module)
                new_modules.append(module)
                continue

            # Check if there is already this module in new_modules, but in
            # different version. If so, raise an exception.
            if module["version"] != old_modules[0]["version"]:
                raise ModuleLookupError(
                    "%s:%s:%s:%s conflicts with %s:%s:%s:%s"
                    % (
                        module["name"],
                        module["stream"],
                        module["version"],
                        module["context"],
                        old_modules[0]["name"],
                        old_modules[0]["stream"],
                        old_modules[0]["version"],
                        old_modules[0]["context"],
                    )
                )

            # Check if there is already this module in new_modules in the very
            # same context - do not add it there, because it would be duplicate.
            if module["context"] in [m["context"] for m in old_modules]:
                continue

            # Add it to new_modules/module_map.
            module_map[key].append(module)
            new_modules.append(module)

        if expand:
            added_module_list = new_modules
            while True:
                added_module_list = self._add_new_dependencies(
                    module_map, added_module_list
                )
                if len(added_module_list) == 0:
                    break
                new_modules.extend(added_module_list)

        return new_modules



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import copy
import six
import flask
from flask import request, url_for, Response
from functools import wraps
from odcs.server.models import Compose
from odcs.server import conf
from odcs.server.errors import Forbidden
from odcs.common.types import (
    COMPOSE_RESULTS,
    COMPOSE_FLAGS,
    INVERSE_PUNGI_SOURCE_TYPE_NAMES,
)


def _set_default_client_allowed_attrs(ret_attrs, attrs):
    """
    Helper method adding default allowed_client_attrs into `ret_attrs`.

    If some requested attributes from `attrs` are missing in `ret_attrs` list,
    try to get them from "conf.allowed_$attr_name". If they are not there
    too, use empty list to disallow everything.
    """
    for attr in attrs:
        if attr not in ret_attrs:
            # If raw_config_keys is not set, it means all the `raw_config_urls` keys
            # are allowed for this client.
            if attr == "raw_config_keys":
                ret_attrs[attr] = conf.raw_config_urls.keys()
            else:
                ret_attrs[attr] = getattr(conf, "allowed_%s" % attr, [])
    return ret_attrs


def _load_allowed_clients_attr(attrs):
    """
    Loads attributes from the conf.allowed_clients dict based on the logged
    in user and its groups. If the requested attribute is not found in
    the loaded dict, the conf.allowed_$attr_name is used as a default.

    :param list attrs: List of attribute names to load from the dict.
    :return: Generator of Dicts with loaded attributes.
    """
    # Check if logged user exists in "users" definition of allowed_clients.
    clients = conf.allowed_clients.get("users", {})
    if flask.g.user.username in clients:
        ret_attrs = dict(copy.deepcopy(clients[flask.g.user.username]))
        ret_attrs = _set_default_client_allowed_attrs(ret_attrs, attrs)
        yield ret_attrs
    else:
        # Check if group is defined in allowed_clients "groups".
        clients = conf.allowed_clients.get("groups", {})
        for group in flask.g.groups:
            if group in clients:
                ret_attrs = dict(copy.deepcopy(clients[group]))
                ret_attrs = _set_default_client_allowed_attrs(ret_attrs, attrs)
                yield ret_attrs


def _enum_int_to_str_list(enum_dict, val):
    """
    Convenient method converting int value to list of strings
    according to enum dict.

    For example for {"a": 1, "b": 2, "c": 4} and val=6, this method
    returns ["b", "c"].
    """
    tmp = []
    for enum_name, enum_val in enum_dict.items():
        if val & enum_val == enum_val:
            tmp.append(enum_name)
    return tmp


def raise_if_input_not_allowed(**kwargs):
    """
    Raises `Forbidden` exception if currently logged in user is not allowed
    to operate with a compose of given attributes passed as kwargs.

    Following attributes are valid:
      - flags - compose.flags int
      - results - copose.reults int
      - source_types - compose.source_type
      - sources - compose.source
      - arches - compose.arches
      - compose_types - compose.compose_type
      - raw_config_keys - compose.source.split("#")[0]

    The decision whether the user is allowed or not is done based on
    conf.allowed_clients value.
    """
    if conf.auth_backend == "noauth":
        return

    errors = set()
    for attrs in _load_allowed_clients_attr(kwargs.keys()):
        found_error = False
        for name, values in kwargs.items():
            if name not in attrs:
                # This should not happen, but be defensive in this part of code...
                errors.add(
                    "User %s not allowed to operate with compose with %s=%r."
                    % (flask.g.user.username, name, values)
                )
                continue

            # Convert integers from db format to string list.
            if name == "source_types":
                values = INVERSE_PUNGI_SOURCE_TYPE_NAMES[values]
            elif name == "flags":
                values = _enum_int_to_str_list(COMPOSE_FLAGS, values)
            elif name == "results":
                values = _enum_int_to_str_list(COMPOSE_RESULTS, values)
            elif name == "target_dirs":
                # The default conf.target_dir is always allowed.
                if values == conf.target_dir:
                    continue
                inverse_extra_target_dirs = {
                    v: k for k, v in conf.extra_target_dirs.items()
                }
                values = inverse_extra_target_dirs[values]

            if type(values) == int:
                values = [values]
            elif isinstance(values, six.string_types):
                # `arches` and `sources` are white-space separated lists.
                values = values.split(" ")
            elif values is None:
                # If value is not set, it means it is not available for this compose
                # type and we need to skip its check later. Set the values to empty
                # list to do that.
                values = []

            for value in values:
                allowed_values = attrs[name]
                if (
                    not allowed_values or value not in allowed_values
                ) and allowed_values != [""]:
                    errors.add(
                        "User %s not allowed to operate with compose with %s=%s."
                        % (flask.g.user.username, name, value)
                    )
                    found_error = True
                    break
        if not found_error:
            return
    if errors:
        raise Forbidden(" ".join(list(errors)))
    else:
        raise Forbidden(
            "User %s not allowed to operate with any compose." % flask.g.user.username
        )


def validate_json_data(dict_or_list, level=0, last_dict_key=None):
    """
    Checks that json data represented by dict `dict_or_list` is valid ODCS
    input. Raises ValueError in case the json data does not pass validation.

    This mainly checks that any value in json data does not contain forbidden
    characters and data types which could potentially lead to dangerous pungi
    configuration being generated.
    """
    if isinstance(dict_or_list, dict):
        iterator = dict_or_list.items()
    else:
        iterator = enumerate(dict_or_list)
    for k, v in iterator:
        if isinstance(v, dict):
            # Allow only dict with "source" key name in first level of
            # json object.
            if level != 0 or k not in ["source"]:
                raise ValueError("Only 'source' key is allowed to contain dict.")
            validate_json_data(v, level + 1, k)
        elif isinstance(v, list):
            validate_json_data(v, level + 1, k)
        elif isinstance(v, six.string_types):
            # Packages are stored in comps.xml, not in pungi.conf, so it is
            # not exploitable.
            if last_dict_key in ["packages"]:
                continue
            allowed_chars = [" ", "-", "/", "_", ".", ":", "#", "+", "?", "$", "~"]
            if not all(c.isalnum() or c in allowed_chars for c in v):
                raise ValueError(
                    "Only alphanumerical characters and %r characters "
                    "are allowed in ODCS input variables" % (allowed_chars)
                )
        elif isinstance(v, (int, float)):
            # Allow int, float and also bool, because that's subclass of int.
            continue
        else:
            raise ValueError(
                "Only dict, list, str, unicode, int, float and bool types "
                "are allowed in ODCS input variables, but '%s' has '%s' "
                "type" % (k, type(v))
            )


def pagination_metadata(p_query, request_args):
    """
    Returns a dictionary containing metadata about the paginated query.
    This must be run as part of a Flask request.
    :param p_query: flask_sqlalchemy.Pagination object
    :param request_args: a dictionary of the arguments that were part of the
    Flask request
    :return: a dictionary containing metadata about the paginated query
    """

    request_args_wo_page = dict(copy.deepcopy(request_args))
    # Remove pagination related args because those are handled elsewhere
    # Also, remove any args that url_for accepts in case the user entered
    # those in
    for key in ["page", "per_page", "endpoint"]:
        if key in request_args_wo_page:
            request_args_wo_page.pop(key)
    for key in request_args:
        if key.startswith("_"):
            request_args_wo_page.pop(key)
    pagination_data = {
        "page": p_query.page,
        "pages": p_query.pages,
        "per_page": p_query.per_page,
        "prev": None,
        "next": None,
        "total": p_query.total,
        "first": url_for(
            request.endpoint,
            page=1,
            per_page=p_query.per_page,
            _external=True,
            **request_args_wo_page
        ),
        "last": url_for(
            request.endpoint,
            page=p_query.pages,
            per_page=p_query.per_page,
            _external=True,
            **request_args_wo_page
        ),
    }

    if p_query.has_prev:
        pagination_data["prev"] = url_for(
            request.endpoint,
            page=p_query.prev_num,
            per_page=p_query.per_page,
            _external=True,
            **request_args_wo_page
        )

    if p_query.has_next:
        pagination_data["next"] = url_for(
            request.endpoint,
            page=p_query.next_num,
            per_page=p_query.per_page,
            _external=True,
            **request_args_wo_page
        )

    return pagination_data


def _order_by(flask_request, query, base_class, allowed_keys, default_key):
    """
    Parses the "order_by" argument from flask_request.args, checks that
    it is allowed for ordering in `allowed_keys` list and sets the ordering
    in the `query`.
    In case "order_by" is not set in flask_request.args, use `default_key`
    instead.

    If "order_by" argument starts with minus sign ('-'), the descending order
    is used.
    """
    order_by = flask_request.args.get("order_by", default_key, type=str)
    if order_by and len(order_by) > 1 and order_by[0] == "-":
        order_asc = False
        order_by = order_by[1:]
    else:
        order_asc = True

    if order_by not in allowed_keys:
        raise ValueError(
            "An invalid order_by key was suplied, allowed keys are: "
            "%r" % allowed_keys
        )

    order_by_attr = getattr(base_class, order_by)
    if not order_asc:
        order_by_attr = order_by_attr.desc()
    return query.order_by(order_by_attr)


def filter_composes(flask_request):
    """
    Returns a flask_sqlalchemy.Pagination object based on the request parameters
    :param request: Flask request object
    :return: flask_sqlalchemy.Pagination
    """
    search_query = dict()

    for key in [
        "owner",
        "source_type",
        "source",
        "state",
        "koji_task_id",
        "pungi_compose_id",
        "compose_type",
        "label",
    ]:
        if flask_request.args.get(key, None):
            search_query[key] = flask_request.args[key]

    query = Compose.query

    if search_query:
        query = query.filter_by(**search_query)

    query = _order_by(
        flask_request,
        query,
        Compose,
        [
            "id",
            "owner",
            "source_Type",
            "koji_event",
            "state",
            "time_to_expire",
            "time_submitted",
            "time_done",
            "time_removed",
        ],
        "-id",
    )

    page = flask_request.args.get("page", 1, type=int)
    per_page = flask_request.args.get("per_page", 10, type=int)
    return query.paginate(page, per_page, False)


def cors_header(allow="*"):
    """
    A decorator that sets the Access-Control-Allow-Origin header to
    the desired value on a Flask route.
    :param allow: a string of the domain to allow. This defaults to '*'.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            rv = func(*args, **kwargs)
            if rv:
                # If a tuple was provided, then the Flask Response should be the first object
                if isinstance(rv, tuple):
                    response = rv[0]
                else:
                    response = rv
                # Make sure we are dealing with a Flask Response object
                if isinstance(response, Response):
                    response.headers.add("Access-Control-Allow-Origin", allow)
            return rv

        return wrapper

    return decorator



"""Store celery task id

Revision ID: a855c39e2a0f
Revises: 82172e6a3154
Create Date: 2020-03-11 08:25:27.973866

"""

# revision identifiers, used by Alembic.
revision = "a855c39e2a0f"
down_revision = "82172e6a3154"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("composes", sa.Column("celery_task_id", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("composes", "celery_task_id")
    # ### end Alembic commands ###
